package games.pandemic;

import core.interfaces.IFeatureRepresentation;
import core.interfaces.IGamePhase;
import core.components.*;
import core.properties.*;
import core.AbstractGameState;
import core.components.Area;
import core.AbstractParameters;
import utilities.Hash;
import utilities.Utils;

import java.util.*;

import static games.pandemic.PandemicConstants.*;
import static core.CoreConstants.*;
import static games.pandemic.PandemicGameState.PandemicGamePhase.Forecast;


public class PandemicGameState extends AbstractGameState implements IFeatureRepresentation {

    // The Pandemic game phase enum distinguishes 3 more phases on top of the default ones for players forced to
    // discard cards, a player wishing to play a "Forecast" event card
    // or an opportunity to play a "Resilient Population" event card.
    public enum PandemicGamePhase implements IGamePhase {
        DiscardReaction,
        Forecast,
        RPReaction
    }

    // Collection of areas, mapped to player ID. -1 is the general game area containing the board, counters and several decks.
    HashMap<Integer, Area> areas;
    // Temporary deck used as a buffer by several actions.
    Deck<Card> tempDeck;

    // The main game board
    GraphBoard world;
    // Was a quiet night card played?
    boolean quietNight;
    // Was an epidemic card drawn?
    boolean epidemic;
    // How many cards the current player has drawn in their turn
    int nCardsDrawn;
    // Keeps track of locations of all research stations (list of names of cities / board nodes)
    ArrayList<String> researchStationLocations;

    @Override
    protected List<Component> _getAllComponents() {
        List<Component> components = new ArrayList<>(areas.values());
        components.add(tempDeck);
        components.add(world);
        return components;
    }

    /**
     * Calculates the list of currently available actions, possibly depending on the game phase.
     * @return - List of AbstractAction objects.
     */
    @Override
    public double[] getDistanceFeatures(int playerId) {
        // Win if all disease counters >= 1
        // Lose if player deck is empty
        // Lose if too many outbreaks
        // Lose if no more disease cubes when needed

        double[] features = new double[3 + colors.length];
        int nDiseasesCured = 0;
        int i = 3;
        for (String color: colors) {
            Counter disease = (Counter) getComponent(Hash.GetInstance().hash("Disease " + color));
            if (disease.getValue() >= 1) nDiseasesCured++;

            Counter diseaseCube = (Counter) getComponent(Hash.GetInstance().hash("Disease cube " + color));
            features[i++] = diseaseCube.getValue();
        }

        int nCardsPlayerDeck = ((Deck<Card>)getComponent(playerDeckHash)).getSize();
        int nOutbreaks = ((Counter)getComponent(outbreaksHash)).getValue();
        features[0] = nDiseasesCured;
        features[1] = nCardsPlayerDeck;
        features[2] = nOutbreaks;
        return features;
    }

    @Override
    public HashMap<HashMap<Integer, Double>, Utils.GameResult> getTerminalFeatures(int playerId) {
        HashMap<HashMap<Integer, Double>, Utils.GameResult> terminals = new HashMap<>();
        terminals.put(new HashMap<Integer, Double>() {{ put(0, (double) colors.length); }}, Utils.GameResult.WIN);
        terminals.put(new HashMap<Integer, Double>() {{ put(1, 0.0); }}, Utils.GameResult.LOSE);
        terminals.put(new HashMap<Integer, Double>() {{ put(2, (double) ((PandemicParameters)gameParameters).lose_max_outbreak); }},
                Utils.GameResult.LOSE);
        int i = 3;
        for (String color: colors) {
            int id = i;
            terminals.put(new HashMap<Integer, Double>() {{ put(id, 0.0); }}, Utils.GameResult.LOSE);
            i++;
        }
        return terminals;
    }


    @Override
    protected double _getScore(int playerId) {
        return new PandemicHeuristic().evaluateState(this, playerId);
    }

    @Override
    protected ArrayList<Integer> _getUnknownComponentsIds(int playerId) {
        return new ArrayList<Integer>() {{
            Deck<Card> pd = (Deck<Card>) getComponent(playerDeckHash);
            Deck<Card> id = (Deck<Card>) getComponent(infectionHash);
            add(pd.getComponentID());
            add(id.getComponentID());
            for (Component c: pd.getComponents()) {
                add(c.getComponentID());
            }
            for (Component c: id.getComponents()) {
                add(c.getComponentID());
            }
        }};
    }

    @Override
    protected void _reset() {
        areas = null;
        tempDeck = null;
        world = null;
        quietNight = false;
        epidemic = false;
        nCardsDrawn = 0;
        researchStationLocations = new ArrayList<>();
    }

    @Override
    protected boolean _equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof PandemicGameState)) return false;
        if (!super.equals(o)) return false;
        PandemicGameState that = (PandemicGameState) o;
        return quietNight == that.quietNight &&
                epidemic == that.epidemic &&
                nCardsDrawn == that.nCardsDrawn &&
                Objects.equals(areas, that.areas) &&
                Objects.equals(tempDeck, that.tempDeck) &&
                Objects.equals(world, that.world) &&
                Objects.equals(researchStationLocations, that.researchStationLocations);
    }

    @Override
    public int hashCode() {
        return Objects.hash(super.hashCode(), areas, tempDeck, world, quietNight, epidemic, nCardsDrawn, researchStationLocations);
    }

    /**
     * Constructor. Calls super with objects corresponding to this game and loads the data for the game.
     * @param pp - Game parameters.
     * @param nPlayers - number of players.
     */
    public PandemicGameState(AbstractParameters pp, int nPlayers) {
        super(pp, new PandemicTurnOrder(nPlayers, ((PandemicParameters)pp).n_actions_per_turn));
        data = new PandemicData();
        data.load(((PandemicParameters)gameParameters).getDataPath());
    }

    // Getters & setters
    public Component getComponent(int componentId, int playerId) {
        return areas.get(playerId).getComponent(componentId);
    }
    public Component getComponentActingPlayer(int componentId) {
        return areas.get(turnOrder.getCurrentPlayer(this)).getComponent(componentId);
    }
    public Component getComponent(int componentId) {
        return getComponent(componentId, -1);
    }
    Area getArea(int playerId) {
        return areas.get(playerId);
    }
    public void addResearchStation(String location) { researchStationLocations.add(location); }
    public void removeResearchStation(String location) { researchStationLocations.remove(location); }
    public void setQuietNight(boolean qn) {
        quietNight = qn;
    }
    public boolean isQuietNight() {
        return quietNight;
    }
    public void setEpidemic(boolean epidemic) {
        this.epidemic = epidemic;
    }
    public boolean isEpidemic() {
        return epidemic;
    }
    public void cardWasDrawn() {
        nCardsDrawn++;
    }
    public void setNCardsDrawn(int nCardsDrawn) {
        this.nCardsDrawn = nCardsDrawn;
    }
    public int getNCardsDrawn() {
        return nCardsDrawn;
    }
    public void clearTempDeck() {
        tempDeck.clear();
    }
    public Deck<Card> getTempDeck() {
        return tempDeck;
    }
    public String getPlayerRoleActingPlayer() {
        return getPlayerRole(turnOrder.getCurrentPlayer(this));

    }
    public String getPlayerRole(int i) {
        Card playerCard = ((Card) getComponent(playerCardHash, i));
        return ((PropertyString) playerCard.getProperty(nameHash)).value;
    }
    public GraphBoard getWorld() {
        return world;
    }

    PandemicData getData() {  // Only FM should have access to this for initialisation
        return (PandemicData)data;
    }

    @Override
    protected AbstractGameState _copy(int playerId) {
        PandemicGameState gs = new PandemicGameState(gameParameters.copy(), getNPlayers());

        gs.areas = new HashMap<>();
        for(int key : areas.keySet())
        {
            Area a = areas.get(key);
            if (playerId != -1 && key == -1) {
                // Hiding face-down decks in game area
                a = new Area(key, "Game area");
                HashMap<Integer, Component> oldComponents = areas.get(key).getComponents();
                for (Map.Entry<Integer, Component> e: oldComponents.entrySet()) {
                    if (PARTIAL_OBSERVABLE && playerId != -1 && (e.getKey() == playerDeckHash || e.getKey() == infectionHash)) {
                        Random r = new Random(gs.getGameParameters().getRandomSeed());
                        Deck<Card> hiddenDeck = (Deck<Card>) e.getValue().copy();
                        if (gamePhase == Forecast && e.getKey() == infectionHash) {
                            // Top N cards should be left the same, the rest shuffled
                            hiddenDeck.shuffle(((PandemicParameters)gameParameters).n_forecast_cards, hiddenDeck.getSize(), r);
                        } else {
                            hiddenDeck.shuffle(r);  // We know what cards are in there, a simple shuffle is enough
                        }
                        a.putComponent(e.getKey(), hiddenDeck);
                    } else {
                        a.putComponent(e.getKey(), e.getValue().copy());
                    }
                }
            }
            gs.areas.put(key, a.copy());
        }
        gs.tempDeck = tempDeck.copy();

        gs.world = world.copy();
        gs.quietNight = quietNight;
        gs.epidemic = epidemic;
        gs.nCardsDrawn = nCardsDrawn;

        gs.researchStationLocations = new ArrayList<>(researchStationLocations);

        return gs;
    }

    void addComponents() {
        super.addAllComponents();
    }
}
